definition module ioTypes


//////////////////////////////////////////////////
//                                               /
// Common types for the event I/O system.        /
//                                               /
// / / / / / / / / / / / / / / / / / / / / / / / /


import StdEnv
import picture
import intrface


  //-------------------------------------------------//
 //  The IO System                                  //
//-------------------------------------------------//


:: IOSystem *s *io :== [ DeviceSystem s io ]

:: DeviceSystem *s *io =  MenuSystem   [MenuDef   s io]
				 		| TimerSystem  [TimerDef  s io]
						| WindowSystem [WindowDef s io]
						| DialogSystem [DialogDef s io]

:: *EVENTS 

:: *IOState *s 


//---------------------------------------------------


:: IOadmin *s = { io_quit            :: !Bool,
			 	  io_menuState       :: !Perhaps (MenuWinAdmin s),
				  io_windowState     :: ![(WindowAdmin s)], 
				  io_timers          :: ![TimerAdmin s],
				  io_dialogs         :: ![DialogAdmin s],
				  io_noticeresult    :: !Perhaps NoticeButtonId,
				  io_delayedmessages :: ![ CrossCallInfo ]
			    }


OpenEvents :: !*World -> (!*EVENTS, !*World)
	

CloseEvents :: !*EVENTS !*World -> *World

PackIOState   :: !(IOadmin s) !OS -> IOState s

UnpackIOState :: !(IOState s) -> (!IOadmin s, !OS)

EventsToOs :: !EVENTS -> OS

OsToEvents :: !OS -> EVENTS

ChangeOs :: !(OS -> OS) !(IOState s) -> (IOState s) 

EmptyIOadmin :: IOadmin s


  //-------------------------------------------------//
 //  Common Defs and simple functions               //
//-------------------------------------------------//




:: ItemTitle     :== String

:: SelectState =  Able | Unable
:: MarkState   =  Mark | NoMark

:: KeyboardState :== (!KeyCode, !KeyState, !Modifiers)
:: KeyCode       :== Char
:: KeyState      =  KeyUp | KeyDown | KeyStillDown

:: MouseState    :== (!MousePosition, !ButtonState, !Modifiers)
:: MousePosition :== (!Int, !Int)
:: ButtonState   =  ButtonUp | ButtonDown | ButtonDoubleDown |
                    ButtonTripleDown | ButtonStillDown

/* Modifiers indicates the meta keys that have been pressed (TRUE)
   or not (FALSE): (Shift, Alternate/Meta, Control, Control)
*/

:: Modifiers     :== (!Bool,!Bool,!Bool,!Bool)

:: PictureDomain :== Rectangle


  //----------------------------------------------//
 //  The Defs for the menu device.               //
//----------------------------------------------//


    
:: MenuDef *s *io
       =  PullDownMenu MenuId MenuTitle SelectState [MenuElement s io]

:: MenuElement *s *io
       =  MenuItem MenuItemId ItemTitle KeyShortcut SelectState (MenuFunction s io)
         |  CheckMenuItem MenuItemId ItemTitle KeyShortcut SelectState MarkState
                          (MenuFunction s io)
         |  SubMenuItem MenuId ItemTitle SelectState [MenuElement s io]
         |  MenuItemGroup MenuItemGroupId [MenuElement s io]
         |  MenuRadioItems MenuItemId [RadioElement s io]
         |  MenuSeparator

:: RadioElement *s *io
       =  MenuRadioItem MenuItemId ItemTitle KeyShortcut SelectState
                        (MenuFunction s io)


:: MenuTitle             :== String
:: MenuFunction *s *io   :== !s -> *( !io -> (s, io)) 
:: MenuId                :== Int
:: MenuItemId            :== Int
:: MenuItemGroupId       :== Int
:: KeyShortcut           =  Key KeyCode | NoKey

//-----------------------------------------------------------------------

:: MenuKind  =  MenuBarK
			  | PullDownK
              | SubMenuK
              | ItemGroupK
              | RadioGroupK

:: MenuItemKind  =  RadioItemK
                  | CheckItemK
                  | ItemK

:: MenuAdmin *s = {  mkind     :: !MenuKind,
				     mid       :: !Int,
				     mtitle    :: !String,
				     melements :: ![MAdmin s],
				     menabled  :: !Bool,
				     mhandle   :: !HMENU,
					 mparent   :: !HMENU,
					 mpos      :: !Int
			       }

:: ItemAdmin *s = {  ikind     :: !MenuItemKind, 
				     iid       :: !Int,
					 ihandle   :: !HITEM,
				     ititle    :: !String,
				     ikey      :: !KeyShortcut,
				     ienabled  :: !Bool,
				     imarked   :: !Bool,
				     ifunction :: !MenuFunction s (IOState s),
				 	 iparent   :: !HMENU,
					 ipos      :: !Int
			       }


:: MenuWinAdmin *s = {  mbwinhandle :: !HWND,
				 	    mbmenu      :: !MAdmin s
				      }

:: MAdmin *s =  Menu !(MenuAdmin s)
              | Item !(ItemAdmin s)  
			  | Separator

  //------------------------------------------------//
 //  The Defs for the window device.               //
//------------------------------------------------//


:: WindowDef * s * io
   =  ScrollWindow WindowId WindowPos WindowTitle
                   ScrollBarDef ScrollBarDef
                   PictureDomain MinimumWindowSize InitialWindowSize
                   (UpdateFunction s) [WindowAttribute s io]
   |  FixedWindow  WindowId WindowPos WindowTitle
                   PictureDomain
                   (UpdateFunction s)  [WindowAttribute s io]
 
:: WindowId          :== Int
:: WindowPos         :== (!Int, !Int)
:: WindowTitle       :== String

:: ScrollBarDef      =  ScrollBar ThumbValue ScrollValue
:: ThumbValue        =  Thumb  Int
:: ScrollValue       =  Scroll Int
:: MinimumWindowSize :== (!Int, !Int)
:: InitialWindowSize :== (!Int, !Int)

:: UpdateFunction * s      :== UpdateArea -> s -> (s,[DrawFunction]) 
:: GoAwayFunction * s * io :== s ->  *(io -> (s, io))
 
:: WindowAttribute * s * io =
      Activate   (WindowFunction s io)
   |  Deactivate (WindowFunction s io)
   |  GoAway     (WindowFunction s io)
   |  Mouse      SelectState (MouseFunction    s io)
   |  Keyboard   SelectState (KeyboardFunction s io)
   |  Cursor     CursorShape
   |  StandByWindow 

:: WindowFunction   * s * io :== s ->  *(io -> (s, io) )
:: KeyboardFunction * s * io :== KeyboardState ->  s ->  *(io -> (s, io))  
:: MouseFunction    * s * io :== MouseState ->     s ->  *(io -> (s, io))  
 

:: CursorShape =  StandardCursor | BusyCursor     | IBeamCursor |
                  CrossCursor    | FatCrossCursor | ArrowCursor | HiddenCursor


//------------------------------------------------------------------


:: WindowKind =  ScrollK
			   | FixedK


:: WindowAdmin *s = {  wkind       :: !WindowKind,
					   wid         :: !Int,
					   whandle     :: !HWND,
			           wtitle      :: !String,
					   wminwidth   :: !Int,
					   wminheight  :: !Int,
					   wupdatef    :: !UpdateFunction s,
				 	   wdomain     :: !PictureDomain,
                       wpicstate   :: !PictureState,
					   whscroll    :: !Int,
					   wvscroll    :: !Int,  
					   whthumb     :: !Int,
					   wvthumb     :: !Int,  
					   wgoaway     :: !WindowFunction s (IOState s),
					   wactivate   :: !WindowFunction s (IOState s),
					   wdeactivate :: !WindowFunction s (IOState s),
					   wkeyboardf  :: !KeyboardFunction s (IOState s),
					   wkeybable   :: !Bool,
					   wmousef     :: !MouseFunction s (IOState s),
					   wmouseable  :: !Bool,
					   wcursor     :: !CursorShape
                    }				
				      
  //----------------------------------------------//
 //  The Defs for the timer device.              //
//----------------------------------------------//


:: TimerDef * s * io		=  Timer TimerId SelectState TimerInterval (TimerFunction s io)
:: TimerId      			:== Int
:: TimerInterval			:== Int
:: TimerFunction * s * io	:== TimerState ->  s -> *( io -> (s,io) ) 
:: TimerState    			:== Int


//-----------------------------------------------------------

:: TimerAdmin *s = { tid       :: !Int,
                     thandle   :: !HITEM,
                     tinterval :: !Int,
					 tlasttime :: !Int,
                     tfunction :: !TimerFunction s (IOState s),
				     table     :: !Bool
			  	   }



  //------------------------------------------------//
 //  The Defs for the dialog device.               //
//------------------------------------------------//


:: * DialogState * s * io 
:: DialogInfo

:: ItemInfo = EditInfo !String
            | RadioGroupInfo !DialogItemId
			| CheckGroupInfo ![ (!DialogItemId, !Bool) ]
			| ControlStateInfo !ControlState 


:: DialogDef * s * io
   =  PropertyDialog DialogId DialogTitle [DialogAttribute] (SetFunction s io)
                     (ResetFunction s io) [DialogItem s io]
   |  CommandDialog  DialogId DialogTitle [DialogAttribute]
                     DialogItemId [DialogItem s io]
   |  AboutDialog    ApplicationName PictureDomain [DrawFunction]
                     (AboutHelpDef s io)

:: ApplicationName :== String
:: AboutHelpDef * s * io = NoHelp
                           | AboutHelp ItemTitle (AboutHelpFunction s io)
:: AboutHelpFunction * s * io :== s ->  *(io -> (s,io)) 

:: DialogAttribute
   =  DialogPos    Measure Measure
   |  DialogSize   Measure Measure
   |  DialogMargin Measure Measure
   |  ItemSpace    Measure Measure
   |  StandByDialog

:: DialogId    :== Int
:: DialogTitle :== String
:: Measure     =   MM Real | Inch Real | Pixel Int
 
:: DialogItem * s * io
   =  DialogButton DialogItemId ItemPos ItemTitle SelectState (ButtonFunction s io)
   |  DialogIconButton DialogItemId ItemPos PictureDomain IconLook SelectState (ButtonFunction s io)
   |  StaticText   DialogItemId ItemPos String
   |  DynamicText  DialogItemId ItemPos TextWidth String
   |  EditText     DialogItemId ItemPos TextWidth NrEditLines String
   |  DialogPopUp  DialogItemId ItemPos SelectState DialogItemId [RadioItemDef s io]
   |  RadioButtons DialogItemId ItemPos RowsOrColumns DialogItemId [RadioItemDef s io]
   |  CheckBoxes   DialogItemId ItemPos RowsOrColumns [CheckBoxDef s io]
   |  Control      DialogItemId ItemPos PictureDomain SelectState ControlState
                       ControlLook ControlFeel (DialogFunction s io)
 
:: DialogItemId  :== Int
:: RowsOrColumns =  Rows Int | Columns Int
:: ItemPos       =  Left | Center | Right        |
                    RightTo DialogItemId         |
                    Below DialogItemId           |
                    XOffset DialogItemId Measure |
                    YOffset DialogItemId Measure |
                    XY Measure Measure           |
                    ItemBox Int Int Int Int
:: IconLook      :== SelectState -> [DrawFunction]
:: TextWidth     :== Measure
:: NrEditLines   :== Int
:: RadioItemDef * s * io =  RadioItem DialogItemId ItemTitle SelectState (DialogFunction s io)
:: CheckBoxDef * s * io  =  CheckBox DialogItemId ItemTitle SelectState MarkState (DialogFunction s io)
 
:: ControlState =  IntCS    Int                       |
                   BoolCS   Bool                      |
                   RealCS   Real                      | 
                   StringCS String                    |
                   PairCS   ControlState ControlState |
                   ListCS   [ControlState]
:: ControlLook :== SelectState ->  ControlState -> [DrawFunction] 
:: ControlFeel :== MouseState ->   ControlState -> (ControlState,[DrawFunction]) 
 
:: SetFunction    * s * io :== ButtonFunction s io
:: ResetFunction  * s * io :== ButtonFunction s io
:: DialogFunction * s * io :== DialogInfo -> (DialogState s io) -> DialogState s io
:: ButtonFunction *s *io :== DialogInfo -> ( s -> * (io -> (s,io)))

 
:: NoticeDef       =  Notice [String] NoticeButtonDef [NoticeButtonDef]
:: NoticeButtonDef =  NoticeButton NoticeButtonId ItemTitle
:: NoticeButtonId  :== Int

//---------------------------------------------------------------------



:: DialogAdmin *s = { dId     :: !Int,
                      dHandle :: !HWND,
					  dKind   :: !DialogKind,
					  dItems  :: ![ControlAdmin s]
					}

:: DialogKind = ModalK | ModelessK

:: ControlAdmin *s = { cId      :: !Int,
                       cHandle  :: !HWND,
					   cKind    :: !ControlKind s,
					   cEnabled :: !Bool,
					   cItems   :: ![ CheckRadioAdmin s ]
				     }

:: ControlKind *s = ButtonK       !(ButtonFunction s (IOState s))
                  | IconButtonK   !PictureState !PictureDomain !IconLook !(ButtonFunction s (IOState s))
			      | StaticTextK
			      | DynamicTextK  
			      | EditTextK     !NrEditLines  !String
			      | PopupK        !DialogItemId	 
			      | RadioButtonsK !DialogItemId 
			      | CheckBoxesK   
			      | CustomK       !PictureState !PictureDomain !ControlState !ControlLook !ControlFeel !(DialogFunction s (IOState s))



:: CheckRadioAdmin *s = { caId      :: !DialogItemId,
                          caHandle  :: !HWND,
                          caKind    :: !CheckRadioKind,
						  caEnabled :: !Bool,
						  caFunct   :: !DialogFunction s (IOState s)
						}

:: CheckRadioKind = RadioK | CheckK !Bool


PackDialogInfo :: ![ !(!DialogItemId, !ItemInfo) ] -> DialogInfo
UnpackDialogInfo :: !DialogInfo -> [ ( !DialogItemId, !ItemInfo) ] 

PackDialogState :: !(DialogAdmin s) !OS -> DialogState s (IOState s)
UnpackDialogState :: !(DialogState s io) -> ( !DialogAdmin s, !OS)

  //------------------------------------------------//
 //      Helper functions                          //
//------------------------------------------------//

CursorShape2Code :: !CursorShape -> Int


SetBetween :: !Int !Int !Int -> Int


MakeMouseState :: !Int !Int !Int !Int -> MouseState

MakeKeyboardState :: !Int !Int !Int -> KeyboardState

toMods :: !Int -> Modifiers


SelectStateEqual :: !SelectState !SelectState -> Bool

MarkEqual  :: !MarkState !MarkState -> Bool

Enabled  :: !SelectState -> Bool

Marked  :: !MarkState -> Bool

:: Id   :== Int
:: Size :== (Int,Int)


:: Perhaps a  =  OK !a
			   | Nope

RectangleDimensions :: !Rectangle -> (Int,Int)

StateMap2 :: ( x s -> s) ![x] !s -> s

Select :: (x -> Bool) x ![x] -> (Bool, x)


// smap :: (a s -> (b,s)) [a] s  -> ([b],s)
smap f as s :== smap_ as s
where
	smap_ []       s  =  ([],s)
	smap_ [a:rest] s  =  ([b:restb],finals)
	where
		(b,s2)         = f a s 
		(restb,finals) = smap_ rest s2

Error :: !String !String !String -> .x
